"use strict";

var commonUtils = require("../../core/utils/common"),
    isDefined = require("../../core/utils/type").isDefined,
    extend = require("../../core/utils/extend").extend,
    stateStoringCore = require("./ui.grid_core.state_storing_core"),
    equalByValue = commonUtils.equalByValue;

// TODO move processLoadState to target modules (data, columns, pagerView)
var processLoadState = function processLoadState(that) {
    var columnsController = that.getController("columns"),
        selectionController = that.getController("selection"),
        exportController = that.getController("export"),
        dataController = that.getController("data"),
        pagerView = that.getView("pagerView");

    if (columnsController) {
        columnsController.columnsChanged.add(function () {
            var columnsState = columnsController.getUserState(),
                columnsStateHash = commonUtils.getKeyHash(columnsState),
                currentColumnsStateHash = commonUtils.getKeyHash(that._state.columns);

            if (!equalByValue(currentColumnsStateHash, columnsStateHash)) {
                extend(that._state, {
                    columns: columnsState
                });

                that.isEnabled() && that.save();
            }
        });
    }

    if (selectionController) {
        selectionController.selectionChanged.add(function (e) {
            extend(that._state, {
                selectedRowKeys: e.selectedRowKeys,
                selectionFilter: e.selectionFilter
            });
            that.isEnabled() && that.save();
        });
    }

    if (dataController) {
        that._initialPageSize = that.option("paging.pageSize");
        dataController.changed.add(function () {
            var userState = dataController.getUserState();

            extend(that._state, userState, {
                allowedPageSizes: pagerView ? pagerView.getPageSizes() : undefined
            });
            that.isEnabled() && that.save();
        });
    }

    if (exportController) {
        exportController.selectionOnlyChanged.add(function () {
            extend(that._state, {
                exportSelectionOnly: exportController.selectionOnly()
            });
            that.isEnabled() && that.save();
        });
    }
};

module.exports = {
    defaultOptions: function defaultOptions() {
        return {
            /**
            * @name GridBaseOptions_stateStoring
            * @publicName stateStoring
            * @type object
            */
            stateStoring: {
                /**
                 * @name GridBaseOptions_stateStoring_enabled
                 * @publicName enabled
                 * @type boolean
                 * @default false
                 */
                enabled: false,
                /**
                 * @name GridBaseOptions_stateStoring_storageKey
                 * @publicName storageKey
                 * @type string
                 * @default null
                 */
                storageKey: null,
                /**
                 * @name GridBaseOptions_stateStoring_type
                 * @publicName type
                 * @type Enums.StateStoringType
                 * @default "localStorage"
                 */
                type: "localStorage",

                /**
                 * @name GridBaseOptions_stateStoring_customLoad
                 * @publicName customLoad
                 * @type function()
                 * @type_function_return Promise<Object>
                 */
                customLoad: null,

                /**
                 * @name GridBaseOptions_stateStoring_customSave
                 * @publicName customSave
                 * @type function(gridState)
                 * @type_function_param1 gridState:object
                 */
                customSave: null,

                /**
                 * @name GridBaseOptions_stateStoring_savingTimeout
                 * @publicName savingTimeout
                 * @type number
                 * @default 2000
                 */
                savingTimeout: 2000
            }
        };
    },
    controllers: {
        stateStoring: stateStoringCore.StateStoringController
    },
    extenders: {
        views: {
            rowsView: {
                init: function init() {
                    var that = this;
                    var dataController = that.getController("data");

                    that.callBase();

                    dataController.stateLoaded.add(function () {
                        if (dataController.isLoaded()) {
                            that.setLoading(false);
                            that.renderNoDataText();
                        }
                    });
                }
            }
        },
        controllers: {
            stateStoring: {
                init: function init() {
                    this.callBase.apply(this, arguments);
                    processLoadState(this);
                },
                /**
                 * @name GridBaseMethods_state
                 * @publicName state()
                 * @return object
                 */
                /**
                 * @name GridBaseMethods_state
                 * @publicName state(state)
                 * @param1 state:object
                 */
                isLoading: function isLoading() {
                    return this.callBase() || this.getController("data").isStateLoading();
                },
                state: function state(_state) {
                    var result = this.callBase.apply(this, arguments);
                    if (_state !== undefined) {
                        this.applyState(extend({}, _state));
                    }
                    return result;
                },
                applyState: function applyState(state) {
                    var that = this,
                        allowedPageSizes = state.allowedPageSizes,
                        searchText = state.searchText,
                        selectedRowKeys = state.selectedRowKeys,
                        selectionFilter = state.selectionFilter,
                        exportController = that.getController("export"),
                        columnsController = that.getController("columns"),
                        scrollingMode = that.option("scrolling.mode");

                    that.component.beginUpdate();

                    if (columnsController) {
                        columnsController.setUserState(state.columns);
                    }

                    if (exportController) {
                        exportController.selectionOnly(state.exportSelectionOnly);
                    }

                    that.option("selectedRowKeys", selectedRowKeys || []);

                    that.option("selectionFilter", selectionFilter);

                    if (allowedPageSizes && that.option("pager.allowedPageSizes") === "auto") {
                        that.option("pager").allowedPageSizes = allowedPageSizes;
                    }

                    that.component.endUpdate();

                    that.option("searchPanel.text", searchText || "");

                    that.option("paging.pageSize", scrollingMode !== "virtual" && scrollingMode !== "infinite" && state.pageSize ? state.pageSize : that._initialPageSize);
                    that.option("paging.pageIndex", state.pageIndex || 0);
                }
            },
            columns: {
                getVisibleColumns: function getVisibleColumns() {
                    var visibleColumns = this.callBase.apply(this, arguments),
                        stateStoringController = this.getController("stateStoring");

                    return stateStoringController.isEnabled() && !stateStoringController.isLoaded() ? [] : visibleColumns;
                }
            },
            data: {
                callbackNames: function callbackNames() {
                    return this.callBase().concat(["stateLoaded"]);
                },
                _refreshDataSource: function _refreshDataSource() {
                    var that = this,
                        callBase = that.callBase,
                        stateStoringController = that.getController("stateStoring");

                    if (stateStoringController.isEnabled() && !stateStoringController.isLoaded()) {
                        clearTimeout(that._restoreStateTimeoutID);

                        that._restoreStateTimeoutID = setTimeout(function () {
                            stateStoringController.load().always(function () {
                                that._restoreStateTimeoutID = null;
                                callBase.call(that);
                                that.stateLoaded.fire();
                            });
                        });
                    } else if (!that.isStateLoading()) {
                        callBase.call(that);
                    }
                },

                isLoading: function isLoading() {
                    var that = this,
                        stateStoringController = that.getController("stateStoring");

                    return this.callBase() || stateStoringController.isLoading();
                },

                isStateLoading: function isStateLoading() {
                    return isDefined(this._restoreStateTimeoutID);
                },

                isLoaded: function isLoaded() {
                    return this.callBase() && !this.isStateLoading();
                },

                dispose: function dispose() {
                    clearTimeout(this._restoreStateTimeoutID);
                    this.callBase();
                }
            }
        }
    }
};